
# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.3'
version = '1.0.3'
full_version = '1.0.3.dev0+8a071c1'
git_revision = '8a071c1cde4e33b2931bd4c870bdc840e902db9e'
release = False

if not release:
    version = full_version
    short_version += ".dev"
